/***************************************************************************
 BioSequence.h  -  description
 -------------------
 begin                : Thu Jun 16 2002
 copyright            : (C) 2002 by Oleksiy Kohany
 email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _BIO_SEQUENCE_H_
#define _BIO_SEQUENCE_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <list>
#include <vector>
#include <iostream>
#include <stdio.h>

#define GAPSYMB '-'
#define STOPCODONSYMB '*'

#define LINELENGTH 70
using namespace std;

#define BASERANGE 128

enum SEQTYPE
{
   SEQTYPE_UNKNOWN, NA, AA
};

//const char sValidASCII[] = "ABCDEFGHIKLMNPQRSTUVWXYZabcdefghiklmnpqrstuvwxyz";

const int MAX_LINELENGTH = 255;

class BioFragment: public pair<int, int>
{
   public:
      string locusName;
      BioFragment(string name, int p1, int p2) : pair<int, int>(p1, p2), locusName(name) {
      }
      BioFragment(const char* name, int p1, int p2) : pair<int, int>(p1, p2), locusName(name) {
      }
};

/** ========================================================== */
class BioSequence: public string
{
   protected:
      string m_sName;
      SEQTYPE seqtype;
      /** ---------------------------------------------------------- */
      void rmTrash();
   public:
      using string::assign;

      BioSequence& operator =(const string& s);
      BioSequence& operator =(const char* s);

      BioSequence& append(const BioSequence& s);
      BioSequence& append(const string& s);
      BioSequence& append(const char* s);
      BioSequence& append(char c);

      BioSequence& assign(const char* s);
      BioSequence& assign(const char* s, size_type n);
      BioSequence& assign(const string& s);
      BioSequence& assign(const string &sSeq, const string &sName);
      BioSequence& assign(const char *pcSeq, const char *pcName);
      /** ---------------------------------------------------------- */
      virtual BioSequence& assign(const BioSequence &Seq);
      /** No descriptions */
      virtual int readHead(istream& is);
      /** No descriptions */
      virtual int readHead(FILE* is);
      /** ---------------------------------------------------------- */
      void formatedPrint(ostream &os, int iLineLen = LINELENGTH) const;
      /** ---------------------------------------------------------- */
      void formatedPrint(FILE *FOUT, int iLineLen = LINELENGTH) const;
      /** ---------------------------------------------------------- */
      void unformatedPrint(ostream &os) const;
      /** ---------------------------------------------------------- */
      void unformatedPrint(FILE *FOUT) const;
      void clear();
      int removeGaps();
      /** ---------------------------------------------------------- */
      BioSequence();
      /** ---------------------------------------------------------- */
      BioSequence(const BioSequence& other);
      /** ---------------------------------------------------------- */
      BioSequence(const string &sSeq, const string &sName);
      /** ---------------------------------------------------------- */
      BioSequence(const char *pcSeq, const char *pcName);
      /** ---------------------------------------------------------- */
      virtual ~BioSequence();
      /** ---------------------------------------------------------- */
      const string& getName() const;
      /** ---------------------------------------------------------- */
      void setName(const string &sName);
      /** ---------------------------------------------------------- */
      void setName(const char *pcName);
      /** ---------------------------------------------------------- */
      static string& complement(string& str);
      /** ---------------------------------------------------------- */
      virtual BioSequence& complement();
      /** ---------------------------------------------------------- */
      BioSequence& operator =(const BioSequence &Seq);
      /** ---------------------------------------------------------- */
      long findBase(long iPos = 0) const;
      /** ---------------------------------------------------------- */
      BioSequence& format();
      /** ---------------------------------------------------------- */
      BioSequence& sqeezeOutGaps();
      /** ---------------------------------------------------------- */
      virtual void printHead(ostream &os) const;
      /** ---------------------------------------------------------- */
      virtual void printHead(FILE* FOUT) const;
      /** ---------------------------------------------------------- */
      virtual void printFASTAHead(ostream &os) const;
      /** ---------------------------------------------------------- */
      virtual void printFASTAHead(FILE* os) const;
      /** ---------------------------------------------------------- */
      virtual void printIGHead(ostream &os) const;
      /** ---------------------------------------------------------- */
      virtual void printIGHead(FILE* os) const;
      /** ---------------------------------------------------------- */
      virtual void print(ostream& os, bool format=true) const;
      /** ---------------------------------------------------------- */
      virtual void print(FILE *FOUT, bool format=true) const;
      /** ---------------------------------------------------------- */
      virtual void printFASTA(ostream& os, bool format=true) const;
      /** ---------------------------------------------------------- */
      virtual void printFASTA(FILE *FOUT, bool format=true) const;
      /** ---------------------------------------------------------- */
      virtual void printIG(ostream& os, bool format=true) const;
      /** ---------------------------------------------------------- */
      virtual void printIG(FILE *FOUT, bool format=true) const;
      /** ---------------------------------------------------------- */
      static char complement(char c);
      /** ---------------------------------------------------------- */
      virtual int read(istream& is) throw();
      /** ---------------------------------------------------------- */
      virtual int read(FILE* is) throw();
      /** ---------------------------------------------------------- */
      BioSequence& toupper();
      /** ---------------------------------------------------------- */
      BioSequence& tolower();
      /** No descriptions */
      SEQTYPE guessType() const;
      /** No descriptions */
      static unsigned int effective_length(const string& s);
      /** No descriptions */
      unsigned int effective_length();
      /** No descriptions */
      int baseCount2position(int n);
      /** Read property of SEQTYPE seqtype. */
      virtual SEQTYPE getseqtype();
      static string& shrink(string& s, char c);
      BioSequence subseq(size_t pos, size_t n) const;
      static string subseq(const string& s, size_t pos, size_t n);
      static vector<char> string2vector(const string& s);
      static string vector2string(const vector<char>& v);
      string::const_iterator next_base(string::const_iterator ptr, int* gaplen = NULL) const;
      string::const_iterator prev_base(string::const_iterator ptr, int* gaplen = NULL) const;
      char next_base(int p, int* gaplen = NULL) const;
      char prev_base(int p, int* gaplen = NULL) const;
      static int baseCount2position(const string& s, int n);
      size_t baseCount(size_t pos, int n) const;
      static size_t baseCount(const string& s, size_t pos, int n);
};

/** ========================================================== */
ostream& operator <<(ostream &os, const BioSequence& seq);

/** ========================================================== */
class BioSubSequence: public BioSequence
{
   protected:
      long m_iFrom;
      long m_iTo;
      string m_sComment;
      list<BioFragment> fragments;
   public:
      using BioSequence::assign;
      using BioSequence::append;
      using BioSequence::operator =;
      BioSubSequence();
      BioSubSequence(const BioSequence& other);
      BioSubSequence(const BioSubSequence& other);
      BioSubSequence(const string &sSeq, const string &sName, long iFrom, long iTo);
      BioSubSequence(const char *pcSeq, const char *pcName, long iFrom, long iTo);
      long getFrom() const;
      long getTo() const;
      const string& getComment() const;
      void setFrom(long iFrom);
      void setTo(long iTo);
      void addComment(const string& sComment);
      void setComment(const string& sComment);
      void setComment(const char *pcComment);
      BioSubSequence& complement();
      virtual void printHead(ostream &os) const;
      virtual void printHead(FILE* FOUT) const;
      virtual void printIGHead(ostream &os) const;
      virtual void printIGHead(FILE* os) const;
      virtual void printFASTAHead(ostream &os) const;
      virtual void printFASTAHead(FILE* os) const;
      virtual BioSubSequence& assign(const BioSequence& sSeq);
      virtual BioSubSequence& assign(const BioSubSequence& sSeq);
      BioSequence& trunc_left(unsigned int pos);
      BioSequence& trunc_right(unsigned int pos);
      BioSubSequence subseq(size_t pos, size_t n, bool literal=false) const;
      void adjust_position(int start, int end);
      virtual int readHead(istream& is);
      virtual int readHead(FILE* is);
      BioSubSequence& append(BioSubSequence& other);
      list<BioFragment>& getFragments();
      void clear();
};
class RatedSequence;
namespace boost {
namespace serialization {
	template<class Archive>
	void serialize(Archive & ar, RatedSequence & s, const unsigned int version);
}
}
class RatedSequence: public BioSequence {
protected:
	double score;
public:
	template<class Archive>
	friend void boost::serialization::serialize(Archive & ar, RatedSequence & s, const unsigned int version);
	RatedSequence();
	RatedSequence(const BioSequence& other);
	RatedSequence(const RatedSequence& other);
	virtual RatedSequence& assign(const RatedSequence& other);
	void setScore(double score);
	double getScore() const;
	void clear();
	virtual void printHead(ostream &os) const;
	virtual void printHead(FILE* FOUT) const;
	virtual void printIGHead(ostream &os) const;
	virtual void printIGHead(FILE* os) const;
	virtual void printFASTAHead(ostream &os) const;
	virtual void printFASTAHead(FILE* os) const;
};

//==========================================================

#endif //_BIO_SEQUENCE_H_
