#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <map>
#include <set>
#include <string>
#include <algorithm>
#include "BioSequence.h"
#include "ExtractFunctions.h"
#include "constants.h"
#include "functions.h"
#include "seqdb.h"
#include <fstream>

using namespace std;

typedef set<string> NAMESET;
//========================================================
long readSubseq(FILE *FIN, long iCount, string& sSeq) {
  char sLine[MAXLINELENGTH];
//  fprintf(stderr, "Adding %d bases\n", iCount);
  while(fgets(sLine, MAXLINELENGTH, FIN) != NULL && sLine[0] != ';' && sLine[0] != '>') {
    for(char *pcLine = sLine; *pcLine != '\n' && *pcLine != '\0'; pcLine++) {
      if(isalpha(*pcLine)) {
        sSeq.append(1, *pcLine);
        if(--iCount <= 0) {
          return 0;
        }
      }
    }
  }
  return iCount;
}
//========================================================
long readSubseq(ifstream& ifs, long iCount, string& sSeq) {
  char sLine[MAXLINELENGTH];
  ifs.clear();
  while(ifs.getline(sLine, MAXLINELENGTH).good() && sLine[0] != ';' && sLine[0] != '>') {
    for(char *pcLine = sLine; *pcLine != '\n' && *pcLine != '\0'; pcLine++) {
      if(isalpha(*pcLine)) {
				sSeq.append(1, *pcLine);
				if(--iCount <= 0) {
					return 0;
				}
      }
    }
  }
  return iCount;
}
//========================================================
int extract(FILE* FIN, long iFrom, long iTo, string& sSeq) {
  long iPos = ftell(FIN);
  long iLineLen = 0;
  long iReadPos = 0;
  char c;
  while((c = fgetc(FIN)) != EOF) {
    if(c == '\n') {
      if(iLineLen > 0) {
				break;
      }
      else {
				cerr << "int extract(FILE*, long, long, string&): Sequence starts with newline\n";
      }
    }
    else {
      iLineLen++;
    }
  }
  iReadPos = iFrom + (iFrom - 1)/iLineLen - 1;
  fseek(FIN, iPos + iReadPos, SEEK_SET);
  return readSubseq(FIN, iTo - iFrom + 1, sSeq);
}
//========================================================
int extract(ifstream* ifs, long iFrom, long iTo, string& sSeq) {
  long iPos = ifs->tellg();
  long iLineLen = 0;
  long iReadPos = 0;
  char c;
  ifs->clear();
  while(ifs->get(c).good()) {
    if(c == '\n') {
      if(iLineLen > 0) {
        break;
      }
      else {
        cerr << "int extract(ifstream*, long, long, string&): Sequence starts with newline\n";
      }
    }
    iLineLen++;
  }
  if(iLineLen == 0) {
    return iTo - iFrom + 1;
  }
  iReadPos = iFrom + (iFrom - 1)/iLineLen - 1;
	ifs->seekg(iPos + iReadPos);
  return readSubseq(*ifs, iTo - iFrom + 1, sSeq);
}
//========================================================
int extract(FILE* FIN, const char* pcName, long iFrom, long iTo, FILE* FOUT) {
  if(iFrom <= 0) {
    iFrom = 1;
  }
  if(iTo <= 0) {
    iTo = 1;
  }
  static BioSubSequence sSeq;
  sSeq.clear();
  sSeq.setName(pcName);
  sSeq.setFrom(min(iFrom, iTo));
  sSeq.setTo(max(iFrom, iTo));
  if(iFrom > iTo) {
    sSeq.setTo(sSeq.getTo() - extract(FIN, iTo, iFrom, sSeq));
    sSeq.complement();
  }
  else {
    sSeq.setTo(sSeq.getTo() - extract(FIN, iFrom, iTo, sSeq));
  }
  sSeq.print(FOUT);
  return sSeq.size();
}
//========================================================
int extract(FILE* ifs, const char* sName, long iFrom, long iTo, BioSubSequence& sSeq) {
  if(iFrom <= 0) {
    iFrom = 1;
  }
  if(iTo <= 0) {
    iTo = 1;
  }
  sSeq.clear();
  sSeq.setName(sName);
  sSeq.setFrom(min(iFrom, iTo));
  sSeq.setTo(max(iFrom, iTo));
  if(iFrom > iTo) {
    sSeq.setTo(sSeq.getTo() - extract(ifs, iTo, iFrom, sSeq));
    sSeq.complement();
  }
  else {
    sSeq.setTo(sSeq.getTo() - extract(ifs, iFrom, iTo, sSeq));
  }
  return sSeq.size();
}
//========================================================
int extract(ifstream* ifs, const char* sName, long iFrom, long iTo, BioSubSequence& sSeq) {
  if(iFrom <= 0) {
    iFrom = 1;
  }
  if(iTo <= 0) {
    iTo = 1;
  }
  sSeq.clear();
  sSeq.setName(sName);
  sSeq.setFrom(min(iFrom, iTo));
  sSeq.setTo(max(iFrom, iTo));
  if(iFrom > iTo) {
    sSeq.setTo(sSeq.getTo() - extract(ifs, iTo, iFrom, sSeq));
    sSeq.complement();
  }
  else {
    sSeq.setTo(sSeq.getTo() - extract(ifs, iFrom, iTo, sSeq));
  }
  return sSeq.size();
}
//========================================================
void cleanRecord(char *pcLine) {
  int iStage = 0;
  for(char *pcPtr = pcLine; *pcPtr != '\0'; pcPtr++) {
    switch(iStage) {
    case 0:
      if(isgraph(*pcPtr)) { iStage++;	}
      break;
    case 1:
      if(isspace(*pcPtr)) { iStage++; }
      break;
    case 2:
      if(isdigit(*pcPtr) || (*pcPtr == '-' && isdigit(*(pcPtr + 1)))) {
	iStage++;
      }
      else {
	*pcPtr = ' ';
      }
      break;
    case 3:
      if(isspace(*pcPtr)) { iStage = 2; }
      break;
    }
  }
}
//========================================================
int loadIndex(FILE *F, POSMAP &pmap) {
  return CSdb::loadIndex(F, pmap);
}
//========================================================
void saveIndex(FILE *F, const POSMAP &pmap) {
  CSdb::saveIndex(F, pmap);
}
//========================================================
FILE* findLocus_File(FILE *INF, const char *pcFName, const char *pLName, POSMAP &pmap) {
  return CSdb::findLocus_File(INF, pcFName, pLName, pmap);
}
//========================================================
FILE* findLocus_Map(const char *pLName, const POSMAP &pmap) {
  return CSdb::findLocus_Map(pLName, pmap);
}
//========================================================
FILE *getLocus(FILE *FIN, const char *pcFName, const char *pLName, POSMAP &pmap) {
  if(FIN == NULL) {
#ifdef DEBUG
    fprintf(stderr, "In map\n");
#endif
    return findLocus_Map(pLName, pmap);
  }
  else {
#ifdef DEBUG
    fprintf(stderr, "In file\n");
#endif
    return findLocus_File(FIN, pcFName, pLName, pmap);
  }
}
//========================================================
int buildIndex(FILE *INF, const char *pcFName, POSMAP &pmap) {
  char line[MAXLINELENGTH];
  char sName[MAXNAMESIZE];
  char *pRes;
  long iPos = 0, iCom = 0;
  rewind(INF);
  *line = 0;
  pmap.clear();
  while(line[0] == ';' || line[0] == '>' || (pRes = fgets(line, MAXLINELENGTH, INF)) != NULL) {
    switch(line[0]) {
    case ';':
      while((pRes = fgets(line, MAXLINELENGTH, INF)) != NULL && line[0] == ';');
      if(pRes == NULL) {
	return 0;
      }
      sscanf(line, "%s", sName);
      break;
    case '>':
      sscanf(line + 1, "%s", sName);
      break;
    default:
      iCom = ftell(INF);
      continue;
    }
#ifdef NOEMPTY
    long iNCom = ftell(INF);
    if(fgets(line, MAXLINELENGTH, INF) != NULL) {
      if(line[0] != '>' && line[0] != ';') {
	iPos = ftell(INF);
	pmap[sName].assign(pcFName, iCom, iPos);
      }
      else {
	iCom = iNCom;
	continue;
      }
    }
    else {
      return pmap.size();
    }
#else
    iPos = ftell(INF);
    pmap[sName].assign(pcFName, iCom, iPos);
#endif //NOEMPTY
    iCom = ftell(INF);
    while(((pRes = fgets(line, MAXLINELENGTH, INF)) != NULL) && (line[0] != ';') && (line[0] != '>')) {
      iCom = ftell(INF);
    }
    if(pRes == NULL) {
      break;
    }
  }
  return pmap.size();
}
//========================================================
void doExtract(FILE *FMAP, FILE *FIN, const char *pcFName, FILE *FOUT, POSMAP &pmap) {
  CLocusPos Pos;
  FILE *F = NULL;
  char sLine[MAXLINELENGTH];
  char sName[MAXNAMESIZE];
  double dFrom, dTo;
  int iLineCnt = 0;
  while(fgets(sLine, MAXLINELENGTH, FMAP) != NULL) {
//    fprintf(stderr, "%s", sLine);
    cleanRecord(sLine);
    iLineCnt++;
    if(sscanf(sLine, "%s%lf%lf", sName, &dFrom, &dTo) < 3) {
      fprintf(stderr, "Check line #%d in map file.\n", iLineCnt);
      fprintf(stderr, "%s\n", sLine);
    }
		if(FIN) {
			 F = findLocus_File(FIN, pcFName, sName, pmap);
		}
		else {
			 F = findLocus_Map(sName, pmap);
		}
    if(F != NULL) {
      extract(F, sName, long(dFrom), long(dTo), FOUT);
    }
    else {
      fprintf(stderr, "Locus %s was not found in input file\n", sName);
    }
    fflush(FOUT);
  }
}
