/*
 * ScoreMatrix.cpp
 *
 *  Created on: Aug 13, 2009
 *      Author: kohany
 */

#include "ScoreMatrix.h"
#include <string.h>
#include <sstream>
#include <stdexcept>
#include <fstream>
#include <iostream>

const string
         ScoreMatrix::ALNMATRIX =
                  "\
GAP -30 -6\n\
  A   R   G   C   Y   T   M   W   S   K   N   X\n\
  9   3  -5 -12 -12 -12   0   0  -9  -9  -1  -3\n\
  3   3   3 -12 -12 -12  -6  -6  -6  -6  -1  -3\n\
 -7   3   9 -12 -12 -12  -9  -9   0   0  -1  -3\n\
-12 -12 -12   9   3  -7   0  -9   0  -9  -1  -3\n\
-12 -12 -12   3   3   3  -6  -6  -6  -6  -1  -3\n\
-12 -12 -12  -5   3   9  -9   0  -9   0  -1  -3\n\
  0  -6  -9   0  -6  -9   0  -6  -6  -9  -1  -3\n\
  0  -6  -9  -9  -6   0  -6   0  -9  -6  -1  -3\n\
 -9  -6   0   0  -6  -9  -6  -9   0  -6  -1  -3\n\
 -9  -6   0  -9  -6   0  -9  -6  -6   0  -1  -3\n\
 -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  -3\n\
 -3  -3  -3  -3  -3  -3  -3  -3  -3  -3  -3  -3\n\
";

//This is a slightly corrected copy of linupmatrix from RepeatModeler by Arian Smith and Robert Hubley
const string
         ScoreMatrix::CONMATRIX =
                 "\
GAP -25 -5\n\
  A   R   G   C   Y   T   K   M   S   W   N   X\n\
  9   0  -8 -15 -16 -17 -13  -3 -11  -4  -5  -7\n\
  2   1   1 -15 -15 -16  -7  -6  -6  -7  -5  -7\n\
 -4   3  10 -14 -14 -15  -2  -9  -2  -9  -5  -7\n\
-15 -14 -14  10   3  -4  -9  -2  -2  -9  -5  -7\n\
-16 -15 -15   1   1   2  -6  -7  -6  -7  -5  -7\n\
-17 -16 -15  -8   0   9  -3 -13 -11  -4  -5  -7\n\
-11  -6  -2 -11  -7  -3  -2 -11  -6  -7  -5  -7\n\
 -3  -7 -11  -2  -6 -11 -11  -2  -6  -7  -5  -7\n\
 -9  -5  -2  -2  -5  -9  -5  -5  -2  -9  -5  -7\n\
 -4  -8 -11 -11  -8  -4  -8  -8 -11  -4  -5  -7\n\
 -5  -5  -5  -5  -5  -5  -5  -5  -5  -5  -1  -7\n\
 -7  -7  -7  -7  -7  -7  -7  -7  -7  -7  -7  -7\n\
";

const char ScoreMatrix::GAPCHAR = '-';
const char ScoreMatrix::STOPCODON = '*';
const char ScoreMatrix::LPADDING = '<';
const char ScoreMatrix::RPADDING = '>';

ScoreMatrix::ScoreMatrix() :
   matrix(NULL), gapIniPenalty(0), gapExtPenalty(0), gapMatchScore(0)
{
   if (!loadDefaultMatrix())
   {
      cerr << "Failed to load default matrix" << endl;
   }
}
ScoreMatrix::ScoreMatrix(const char* file) :
   matrix(NULL), gapIniPenalty(0), gapExtPenalty(0), gapMatchScore(0)
{
   if (!parse(file))
   {
      if (!loadDefaultMatrix())
      {
         cerr << "Failed to load default matrix" << endl;
      }
   }
}
ScoreMatrix::ScoreMatrix(const ScoreMatrix& other)
{
   assign(other);
}
ScoreMatrix::~ScoreMatrix()
{
   if (matrix != NULL)
   {
      delete[] matrix;
   }
}
ScoreMatrix& ScoreMatrix::operator =(const ScoreMatrix& other)
{
   return assign(other);
}
ScoreMatrix& ScoreMatrix::assign(const ScoreMatrix& other)
{
   if (matrix != NULL)
   {
      delete[] matrix;
   }
   alphabethSize = other.alphabethSize;
   matrix = new double[alphabethSize * alphabethSize];
   memcpy(matrix, other.matrix, alphabethSize * alphabethSize * sizeof(double));
   memcpy(char2index, other.char2index, sizeof(other.char2index));
   memcpy(index2char, other.index2char, sizeof(index2char));
   gapIniPenalty = other.gapIniPenalty;
   gapExtPenalty = other.gapExtPenalty;
   gapMatchScore = other.gapMatchScore;
   return *this;
}
string& ScoreMatrix::trim(string& s, char c)
{
   size_t pos = 0;
   while (s[pos] == c)
   {
      pos++;
   }
   s.erase(0, pos);
   return s;
}
bool ScoreMatrix::loadDefaultMatrix(const string& defaultMatrix)
{
   istringstream iss(defaultMatrix);
   return parse(iss);
}
bool ScoreMatrix::parse(const char* file)
{
   ifstream ifs;
   ifs.open(file);
   if (ifs.good())
   {
      bool rv = parse(ifs);
      ifs.close();
      return rv;
   }
   else
   {
      if(file[0] != '\0') {
         cerr << "ERROR: Failed to read " << file << endl;
      }
   }
   return false;
}

/**  */
bool ScoreMatrix::parse(istream& is)
{
   double *newMatrix = NULL;
   double *rowptr;
   short newChar2Index[255];
   char newIndex2Char[255];
   memset(newChar2Index, -1, sizeof(newChar2Index));
   memset(newIndex2Char, 0, sizeof(newIndex2Char));
   string line;
   int gi = gapIniPenalty;
   int gx = gapExtPenalty;
   size_t pos;
   int size = 0;
   int index = 0;
   istringstream iss;

   try
   {
      while (!getline(is, line).fail())
      {
         char c;
         string s;
         pos = line.find('#');
         if (pos < line.size())
         {
            if (pos == 0)
            {
               continue;
            }
            line.erase(pos);
         }
         trim(trim(line, ' '), '\t');
         if (!isgraph(line[0]))
         {
            continue;
         }
         if (line.find("FREQS") == 0)
         {
            continue;
         }
         pos = line.find("GAP");
         /**
          * Reading GAP penalties if available
          */
         if (pos < line.size())
         {
            s = line.substr(pos + 3);
            trim(s);
            iss.str(s);
            iss.clear();
            iss >> gi >> gx;
            if (!iss.fail())
            {
               gapIniPenalty = gi;
               gapExtPenalty = gx;
               gapMatchScore = -gx/2;
            }
            else
            {
               cerr << "ERROR: " << line << endl;
               cerr << "ERROR: failed to read gap penalties" << endl;
            }
            continue;
         }
         if (isalpha(line[0]) && size == 0)
         {
            iss.str(line);
            iss.clear();
            for (iss >> s; !iss.fail(); iss >> s)
            {
               if (s.size() > 1)
               {
                  throw logic_error("Expected single character in score matrix column header");
               }
               c = s[0];
               newChar2Index[tolower(c)] = size;
               newChar2Index[toupper(c)] = size;
               newIndex2Char[size] = c;
               size++;
            }
            if (size > 0)
            {
               newMatrix = new double[size * size];
               memset(newMatrix, 0, sizeof(double) * size * size);
               rowptr = newMatrix;
            }
            continue;
         }
         if (size > 0)
         {
            iss.str(line);
            iss.clear();
            if (isalpha(line[0]))
            {
               iss >> s;
               if (s.size() > 1)
               {
                  throw logic_error("Expected single character in score matrix row header");
               }
               c = s[0];
               if (newChar2Index[c] == -1)
               {
                  throw logic_error("Row index character is not present in column index");
               }
               index = newChar2Index[c];
            }
            rowptr = newMatrix + size * index;
            index++;
            int v;
            int rowIndex = 0;
            for (iss >> v; !iss.fail(); iss >> v)
            {
               rowptr[rowIndex++] = v;
            }
            if (rowIndex < size)
            {
               cerr << "Read fever than expected values from score matrix's row: " << rowIndex << "<" << size << endl;
               cerr << "line parsed: " << line << endl;
               throw logic_error("Read fever than expected values from score matrix's row");
            }
         }
      }
      alphabethSize = size;
      memcpy(char2index, newChar2Index, sizeof(newChar2Index));
      memcpy(index2char, newIndex2Char, sizeof(newIndex2Char));
      if (matrix != NULL)
      {
         delete[] matrix;
      }
      matrix = newMatrix;
      return true;
   }
   catch (logic_error exc)
   {
      if (newMatrix != NULL)
      {
         delete[] newMatrix;
      }
      cerr << exc.what() << " at " << __FILE__ << ":" << __LINE__ << endl;
      return false;
   }
}

void ScoreMatrix::print(ostream& os) const
{
   if (gapIniPenalty * gapExtPenalty != 0)
   {
      os << "GAP " << gapIniPenalty << " " << gapExtPenalty << endl;
   }
   for (int i = 0; i < alphabethSize; i++)
   {
      os << "\t" << index2char[i];
   }
   os << endl;
   for (int y = 0; y < alphabethSize; y++)
   {
      os << index2char[y];
      for (int x = 0; x < alphabethSize; x++)
      {
         os << "\t" << matrix[y * alphabethSize + x];
      }
      os << endl;
   }
}

double ScoreMatrix::getScore(char c1, char c2) const
{
   if (c1 == LPADDING || c2 == LPADDING || c1 == RPADDING || c2 == RPADDING) {
      return 0;
   }
   if (c1 == GAPCHAR || c2 == GAPCHAR)
   {
      if (c1 != GAPCHAR || c2 != GAPCHAR)
      {
         return gapExtPenalty;
      }
      return gapMatchScore;
   }
   if (char2index[c1] < 0)
   {
      cerr << "ERROR: Scores undefined for specified symbol c1='" << c1 << "' [" << (int)c1 << "]" << endl;
      throw logic_error("ERROR: Score undefined for specified symbol");
   }
   if (char2index[c2] < 0)
   {
      cerr << "ERROR: Scores undefined for specified symbol c2='" << c2 << "' [" << (int)c2 << "]" << endl;
      throw logic_error("ERROR: Score undefined for specified symbol");
   }
   return matrix[char2index[c1] * alphabethSize + char2index[c2]];
}

string ScoreMatrix::getAlphabeth() const
{
   string alphabeth(alphabethSize, '\0');
   for (int i = 0; i < alphabethSize; i++)
   {
      alphabeth[i] = index2char[i];
   }
   return alphabeth;
}
double ScoreMatrix::getMaxScore() const
{
   if (matrix == NULL)
   {
      return 0;
   }
   double maxWeight = matrix[0];
   for (int i = 1; i < alphabethSize; i++)
   {
      if (maxWeight < matrix[i * alphabethSize + i])
      {
         maxWeight = matrix[i * alphabethSize + i];
      }
   }
   return maxWeight;
}
double ScoreMatrix::getGapIniPenalty() const
{
   return gapIniPenalty;
}
double ScoreMatrix::getGapExtPenalty() const
{
   return gapExtPenalty;
}
double ScoreMatrix::getGapMatchScore() const {
   return gapMatchScore;
}
void ScoreMatrix::setGapIniPenalty(double penalty)
{
   gapIniPenalty = penalty;
}
void ScoreMatrix::setGapExtPenalty(double penalty)
{
   gapExtPenalty = penalty;
}
void ScoreMatrix::setGapMatchScore(double score) {
   gapMatchScore = score;
}
