/***************************************************************************
                          mapcat.h  -  description
                             -------------------
    begin                : Fri Dec  6 15:19:06 PST 2002
    copyright            : (C) 2002 by Oleksiy Kohany
    email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _MAPCAT_H_
#define _MAPCAT_H_

#include <stdio.h>
#include "constants.h"
#include "segment.h"
#include <map>
#include <list>
#include <string>

using namespace std;

class CMatch;
typedef list<CMatch> MATCHLIST;
typedef map<string, MATCHLIST>REPMAP;

const int MAXHOLE = 30;

enum CMPVALS { SCORE, SIM, REPFROM, REPTO, TGTFROM, TGTTO };

class CMatch {
  double m_dScore;
  double m_dMatchRate;
  double m_dTranzRate;
  bool m_bIsDirect;
  char m_sRepName[MAXNAMESIZE];
  char m_sTgtName[MAXNAMESIZE];
  segment repSeg;
  segment tgtSeg;
protected:
//--------------------------------------
  long getCmpVal();
  bool arrange(MATCHLIST &list, MATCHLIST::iterator &lptr);
public:
  static CMPVALS CMPBY;
//--------------------------------------
  static void compareBy(CMPVALS V);
//--------------------------------------
  CMatch() {
    m_dTranzRate = m_dMatchRate = m_dScore = 0;
    m_sRepName[0] = m_sTgtName[0] = '\0';
  }
//--------------------------------------
  CMatch(char *pRepName, int iRepFrom, int iRepTo, 
         char *pTgtName, int iTgtFrom, int iTgtTo, 
         bool isDirect, double dScore, double dMatchRate, double dTranzRate) {
    assign(pRepName, iRepFrom, iRepTo,  pTgtName, iTgtFrom, iTgtTo, isDirect, dScore, dMatchRate, dTranzRate);
  }
//--------------------------------------
  bool isDirect();
//--------------------------------------
  void assign(const char *pRepName, const int iRepFrom, const int iRepTo,
              const char *pTgtName, const int iTgtFrom, const int iTgtTo,
              const bool isDirect, const double dScore, const double dMatchRate, const double dTranzRate);
//--------------------------------------
  void assign(CMatch another);
//--------------------------------------
  const segment& getRepSeg() const;
//--------------------------------------
  const segment& getTgtSeg() const;
//--------------------------------------
  const char *getRepName() const;
//--------------------------------------
  const char *getTgtName() const;
//--------------------------------------
  const int getRepFrom() const;
//--------------------------------------
  const int getTgtFrom() const;
//--------------------------------------
  const int getRepTo()const;
//--------------------------------------
  const int getTgtTo() const;
//--------------------------------------
  const double getScore() const;
//--------------------------------------
  const double getMatchRate() const;
//--------------------------------------
  const double getTranzRate() const;
//--------------------------------------
  void setRepFrom(int iRepFrom);
//--------------------------------------
  void setRepTo(int iRepTo);
//--------------------------------------
  void setTgtFrom(int iTgtFrom);
//--------------------------------------
  void setTgtTo(int iTgtTo);
//--------------------------------------
  int repLength();
//--------------------------------------
  int tgtLength();
//--------------------------------------
  double getMatchCount();
//--------------------------------------
  bool closeTo(CMatch &match, bool isQUp=false, bool isDBp=false, double dCloseRate = 0.15, int iMaxHole = MAXHOLE);
//--------------------------------------
  bool overlaps(CMatch &match, double dOvlpRate = 0.7);
//--------------------------------------
  void join(CMatch &match);
//--------------------------------------
  void joinSimple(CMatch &match);
//--------------------------------------
  void extendBy(CMatch &match);
//--------------------------------------
  void eraseBy(CMatch &match);
//--------------------------------------
  void print(char *pcLine);
//--------------------------------------
  void print(FILE *FOUT);
//--------------------------------------
  bool operator == (CMatch &match);
//--------------------------------------
  bool operator > (CMatch &match);
//--------------------------------------
  bool operator < (CMatch &match);
//--------------------------------------
  bool operator >= (CMatch &match);
//--------------------------------------
  bool operator <= (CMatch &match);
  /** No descriptions */
  void print(ostream* os);
  /** No descriptions */
  int read(char* s);
  /** No descriptions */
  bool isSelfMatch() const;
};
//*******************************************
void sortMap(REPMAP &rmap);
//*******************************************
int loadMap(FILE *FIN, REPMAP &rmap, bool noSelf=false);
//*******************************************
void printMap(FILE *FOUT, REPMAP &rmap);
//*******************************************
bool isAlu(const char *pcRepName);
//*******************************************
int catMapX(REPMAP &rmap, double iOvlpRate = 0.8, bool doCombine = true, bool isQUp = false, bool isDBp = false, int iMinLen = 21, float dMinSim = 0.6);
ostream& operator << (ostream& os, CMatch& mt);
#endif // _MAPCAT_H_
