/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Wed Jan  8 11:03:45 UTC 2003
    copyright            : (C) 2003 by Oleksiy Kohany
    email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <fstream>
#include <map>
#include <list>
#include <stdlib.h>
#include <stdio.h>
#include <exception>
#include <unistd.h>
#include <string.h>
#include "cmaprecord.h"
#include "functions.h"
#include "incompletedata_exception.h"

using namespace std;

char MASK_CHAR = 'N';

typedef list<CMapRecord> MLIST;
typedef map<string, MLIST> MMAP;

void sortMap(MMAP& mmap);
void maskMap(MMAP& mmap, fstream* fs);
void maskMap(MMAP& mmap, FILE* fs);
void maskMap(MMAP& mmap, int filedes);
//====================================================================
#ifdef IOSTREAM
int loadMap(MMAP& mmap, istream *is) {
#else
int loadMap(MMAP& mmap, FILE *is) {
#endif
  CMapRecord rc;
  try {
    while(rc.read(is)) {
      mmap[rc.getTGT()].push_back(rc);
    }
  }
  catch(IncompleteDataException& exc) {
    cerr << exc.what();
  }
  catch(...) {
    cerr << "Bad exception\n";
    throw new bad_exception;
  }
  sortMap(mmap);
  return mmap.size();
}
//====================================================================
void sortMap(MMAP& mmap) {
	MMAP::iterator ptr;
	CMapRecord::setcmpby(TF);
	for(ptr = mmap.begin(); ptr != mmap.end(); ptr++) {
		(*ptr).second.sort();
	}
}
//====================================================================
#ifdef IOSTREAM
void printMap(MMAP& mmap, ostream* os) {
#else
void printMap(MMAP& mmap, FILE* os) {
#endif
  MMAP::iterator mptr;
  MLIST::iterator lptr;
  CMapRecord::setcmpby(TF);
  for(mptr = mmap.begin(); mptr != mmap.end(); mptr++) {
    for(lptr = (*mptr).second.begin(); lptr != (*mptr).second.end(); lptr++) {
      (*lptr).print(os);
    }
  }
}
/** ========= Extremely slow, but should work on all systems ========= */
void maskMap(MMAP& mmap, fstream* fs) {
   char sLine[MAXLINELENGTH];
   char sName[MAXNAMESIZE];
   unsigned long tPos = 0;
   char c;
   MMAP::iterator mptr;
   MLIST::iterator lptr;
   sLine[0] = '\0';
   while(sLine[0] == ';' || sLine[0] == '>' || !fs->getline(sLine, MAXLINELENGTH).fail()) {
      switch(sLine[0]) {
         case '>':
         sscanf(sLine + 1, "%s", sName);
         tPos = 0;
         break;
         case ';':
         while(!fs->getline(sLine, MAXLINELENGTH).eof() && sLine[0] == ';' );
         sscanf(sLine, "%s", sName);
         tPos = 0;
         break;
      }
      sLine[0] = '\0';
      if((mptr = mmap.find(sName)) == mmap.end()) {
         continue;
      }
      lptr = (*mptr).second.begin();




      while(((*fs) >> c).good()) {
         if(c == ';' || c == '>') {
            fs->unget();
            break;
         }
//         if(isalpha(c)) {
         if(isgraph(c)) {
            tPos++;
            if(c != MASK_CHAR) {
               if(tPos > (*lptr).getTGTTo()) {
                  while (lptr != (*mptr).second.end() && tPos >= (*lptr).getTGTTo()) {
                     lptr++;
                  }
                  if(lptr == (*mptr).second.end()) {
                     break;
                  }
               }
               if(tPos >= (*lptr).getTGTFrom() && tPos <= (*lptr).getTGTTo()) {
                  fs->seekp(static_cast<std::streamoff> (-1), ios::cur);
                  (*fs) << MASK_CHAR;
               }
            }
         }
      }
   }
}
#if 0
//====================================================================  
void maskMap(MMAP& mmap, FILE* fs) {  
   int fd = fileno(fs);  
   maskMap(mmap, fd); 
} 
#else 
/** ======== This one is the fastest among those which take READ-WRITE stream as input,
but on Solaris 8(9) using fseek does unpredictable things. On Linux it works fine. ======= */
void maskMap(MMAP& mmap, FILE* fs) {
   char sLine[MAXLINELENGTH];
   char sName[MAXNAMESIZE];
   unsigned long tPos = 0;
   char c;
   MMAP::iterator mptr;
   MLIST::iterator lptr;
   sLine[0] = '\0';
   while(sLine[0] == ';' || sLine[0] == '>' || fgets(sLine, MAXLINELENGTH, fs) != 0) {
      switch(sLine[0]) {
         case '>':
         sscanf(sLine + 1, "%s", sName);
         tPos = 0;
         break;
         case ';':
         while(fgets(sLine, MAXLINELENGTH, fs) != 0 && sLine[0] == ';' );
         sscanf(sLine, "%s", sName);
         tPos = 0;
         break;
      }
      sLine[0] = '\0';
      if((mptr = mmap.find(sName)) == mmap.end()) {
         continue;
      }
      lptr = (*mptr).second.begin();
      while((c = fgetc(fs)) != EOF) {
         if(c == ';' || c == '>') {
            ungetc(c, fs);
            break;
         }
//         if(isalpha(c)) {
         if(isgraph(c)) {
            tPos++;
            if(c != MASK_CHAR) {
               if(tPos > (*lptr).getTGTTo()) {
                  while (lptr != (*mptr).second.end() && tPos >= (*lptr).getTGTTo()) {
                     lptr++;
                  }
                  if(lptr == (*mptr).second.end()) {
                     break;
                  }
               }
               if(tPos >= (*lptr).getTGTFrom() && tPos <= (*lptr).getTGTTo()) {
                  fseek(fs, -1, SEEK_CUR); 
                  fputc(MASK_CHAR, fs);
               }
            }
         }
      }
   }
}
#endif
//====================================================================
int getline(char *buff, int size, int fd) {
   int rdsz = 0;
   while(read(fd, buff, 1)) {
      if(*buff++ == '\n') { break; }
      if(++rdsz >= size) { break; }
   }
   return rdsz;
}
/** ======== This version of maskMap works fine on all tested systems but 
slower than previous one ======== */
void maskMap(MMAP& mmap, int fd, bool lc=false) {
   char sLine[MAXLINELENGTH];
   char sName[MAXNAMESIZE];
   unsigned long tPos = 0;
   char c;
   MMAP::iterator mptr;
   MLIST::iterator lptr;
   sLine[0] = '\0';
   while(sLine[0] == ';' || sLine[0] == '>' || getline(sLine, MAXLINELENGTH, fd) != 0) {
      switch(sLine[0]) {
         case '>':
         sscanf(sLine + 1, "%s", sName);
         tPos = 0;
         break;
         case ';':
         while(getline(sLine, MAXLINELENGTH, fd) != 0 && sLine[0] == ';' );
         sscanf(sLine, "%s", sName);
         tPos = 0;
         break;
      }
      sLine[0] = '\0';
      if((mptr = mmap.find(sName)) == mmap.end()) {
         continue;
      }
      lptr = (*mptr).second.begin();
      while(read(fd, &c, 1)) {
         if(c == ';' || c == '>') {
            lseek(fd, -1, SEEK_CUR);
            break;
         }
//         if(isalpha(c)) {
         if(isgraph(c)) {
            tPos++;
            if(c != MASK_CHAR) {
               if(tPos > (*lptr).getTGTTo()) {
                  while (lptr != (*mptr).second.end() && tPos >= (*lptr).getTGTTo()) {
                     lptr++;
                  }
                  if(lptr == (*mptr).second.end()) {
                     break;
                  }
               }
               if(tPos >= (*lptr).getTGTFrom() && tPos <= (*lptr).getTGTTo()) {
                  lseek(fd, -1, SEEK_CUR);
                  c = tolower(c);
                  lc ? write(fd, &c, 1) : write(fd, &MASK_CHAR, 1);
               }
            }
         }
      }
   }
}
/** ==== This one is the fastest, but it needs output file to be different than input==== */
void maskMap(MMAP& mmap, FILE* ifs, FILE* ofs, bool lc=false) {
   string sLine;
   char sName[MAXNAMESIZE];
   unsigned long tPos = 0;
   char c;
   MMAP::iterator mptr;
   MLIST::iterator lptr;
//   while(sLine[0] == ';' || sLine[0] == '>' || fgets(sLine, MAXLINELENGTH, ifs) != 0) {
   while(sLine[0] == ';' || sLine[0] == '>' || getline(ifs, sLine) != 0) {
       switch(sLine[0]) {
         case '>':
         sscanf(sLine.c_str() + 1, "%s", sName);
         tPos = 0;
         break;
         case ';':
         do {
//            fputs(sLine.c_str(), ofs);
           fprintf(ofs, "%s\n", sLine.c_str());
//         } while(fgets(sLine, MAXLINELENGTH, ifs) != 0 && sLine[0] == ';' );
         } while(getline(ifs, sLine) != 0 && sLine[0] == ';' );
         sscanf(sLine.c_str(), "%s", sName);
         tPos = 0;
         break;
         default:
            if(lc) { toUpper(sLine); }
      }
      fprintf(ofs, "%s\n", sLine.c_str());
      sLine = "";
      if((mptr = mmap.find(sName)) == mmap.end()) {
         continue;
      }
      lptr = (*mptr).second.begin();
      while((c = fgetc(ifs)) != EOF) {
         if(c == ';' || c == '>') {
            ungetc(c, ifs);
            break;
         }
//         if(isalpha(c)) {
         if(isgraph(c)) {
            tPos++;
            if(tPos > (*lptr).getTGTTo()) {
               while (lptr != (*mptr).second.end() && tPos >= (*lptr).getTGTTo()) {
                  lptr++;
               }
               if(lptr == (*mptr).second.end()) {
                  lc ? fputc(toupper(c), ofs) : fputc(c, ofs);
                  break;
               }
            }
            if(tPos >= (*lptr).getTGTFrom() && tPos <= (*lptr).getTGTTo()) {
               lc ? fputc(tolower(c), ofs) : fputc(MASK_CHAR, ofs);
            }
            else {
               lc ? fputc(toupper(c), ofs) : fputc(c, ofs);
            }
         }
         else {
            fputc(c, ofs);
         }
      }
   }
}
#if 0
//====================================================================
void usage(const char* exec) {
   cerr << "\nNAME\n\t" << exec;
#ifdef VERSION
   cerr << " V" << VERSION << "\n";
#endif
   cerr << "\nSYNOPSYS\n\t" << exec << " [inmap] infile [outfile]\n";
   cerr << "\nDESCRIPTION\n\tMasks out repeats. Sequences with repeats masked with '" << MASK_CHAR <<
   "'\n\twill be saved in outfile. infile and outfile can be the same file.\n";
}
//====================================================================
int main(int argc, char *argv[])
{
#ifdef IOSTREAM
   fstream *fs;
   istream *is;
#else
   FILE *fs;
   FILE *is;
#endif
   char **args = argv;
   char *outname = 0;
   if((args = assigninput(is, args)) ==  NULL) {
		usage(argv[0]);
		exit(10);
	}
	if(args[1] == 0) {
		usage(argv[0]);
		exit(20);
	}
	if(args[2] == 0 || strcmp(args[1], args[2]) == 0) {
		outname = args[1];
	}
	else {
		char cmd[MAXLINELENGTH];
		strcpy(cmd, "/bin/cp ");
		strcat(cmd, args[1]);
		strcat(cmd, " ");
		strcat(cmd, args[2]);
		if(system(cmd) != 0) {
			cerr << cmd << " failed.\n";
			exit(30);
		}
		outname = args[2];
	}

#ifdef IOSTREAM
  fs = new fstream(outname);
  if(!fs->good()) {
    cerr << "Failed to open " << outname << " for reading and writing\n";
    exit(40);
  }
#else
  if((fs = fopen(outname, "r+")) == 0) {
    cerr << "Failed to open " << outname << " for reading and writing\n";
    exit(40);
  }
#endif
//===================================================================
  MMAP mmap;
  loadMap(mmap, is);
  maskMap(mmap, fs);
//===================================================================
#ifdef IOSTREAM
  fs->close();
  delete fs;
  delete is;
#else
  fclose(fs);
#endif
  return EXIT_SUCCESS;
}
#else
//====================================================================

void usage(const char* exec) {
   if(!exec) {
      return;
   }
   if(strrchr(exec, '/')) {
      exec = strrchr(exec, '/') + 1;
   }
   cerr << "NAME\n\t" << exec;
#ifdef VERSION
   cerr << " V" << VERSION;
#endif
   cerr << " - masks out fragments from genomic sequences\n";
   cerr << "\nSYNOPSIS\n\t" << exec << " [inmap] infile [outfile] [-c MASKCHAR]\n";
   cerr << "\nDESCRIPTION\n\tMasks out fragments from genomic sequences.\n";
   cerr << "\tMasked sequences will be saved in outfile.\n";
   cerr << "OPTIONS\n\t-c\n\t   sets masking symbol. Default is '" << MASK_CHAR << "'\n";
   cerr << "FLAGS\n\t-lc\n\t   activates lowercase masking\n";
#ifdef COPYRIGHT
   cerr << "\nCOPYRIGHT\n\t" << COPYRIGHT << "\n";
#endif
}
//====================================================================
int main(int argc, char *argv[])
{
#ifdef IOSTREAM
   ostream *mfs;
   ifstream *ifs;

   ostream *ofs;
#else
   FILE *mfs;
   FILE *ifs;

   FILE *ofs;
#endif
   char **args = argv;
   char *outname = 0;
   ARGMAP amap;
   bool bLC = 0;
   getoptions(argv, amap, 1, "-c");
   getflags(argv, amap, 3, "-lc", "--help", "--version");
   if(amap.find("-c") != amap.end()) {
      sscanf(amap["-c"].c_str(), "%c", &MASK_CHAR);
   }
   if(amap.find("--help") != amap.end() || amap.find("--version") != amap.end()) {
      usage(argv[0]);
      exit(0);
   }
   if(amap.find("-lc") != amap.end()) {
     bLC = 1;
   }
   if((args = assigninput(mfs, args)) ==  0) {
      usage(argv[0]);
      exit(10);
   }
   if(args[1] == 0) {
      usage(argv[0]);
      exit(20);
   }
#ifdef IOSTREAM
   if((ifs = openifs(args[1])) == 0) {
      cerr << "Failed to open " << args[1] << "\n";
      usage(argv[0]);
      exit(30);
   }
#else
   if((ifs = openifilename(args[1])) == 0) {
      cerr << "Failed to open " << args[1] << "\n";
      usage(argv[0]);
      exit(40);
   }
#endif
   args++;
   if((args = assignoutput(ofs, args)) ==  0) {
      usage(argv[0]);
      exit(50);
   }
//===================================================================
  MMAP mmap;
  loadMap(mmap, mfs);
  maskMap(mmap, ifs, ofs, bLC);
//===================================================================
#ifdef IOSTREAM
  ifs->close();
  ofs->flush();
  delete ifs;
  delete ofs;
#else
  fclose(ifs);
  fflush(ofs);
#endif
  return EXIT_SUCCESS;
}
#endif
